@@ -1,84 +1,78 @@
+from __future__ import annotations
 from enum import IntEnum, IntFlag
-from typing import List, TypeVar, Tuple, BinaryIO
+from typing import List, Tuple, BinaryIO
 from dataclasses import dataclass
 import struct
 import colorsys
 
 HEADER_SIZE = 16
 
+CONNECTION_ERRORS = (ConnectionResetError, BrokenPipeError, TimeoutError)
+
 
 class ModeFlags(IntFlag):
-    MODE_FLAG_HAS_SPEED = (1 << 0)
-    MODE_FLAG_HAS_DIRECTION_LR = (1 << 1)
-    MODE_FLAG_HAS_DIRECTION_UD = (1 << 2)
-    MODE_FLAG_HAS_DIRECTION_HV = (1 << 3)
-    MODE_FLAG_HAS_BRIGHTNESS = (1 << 4)
-    MODE_FLAG_HAS_PER_LED_COLOR = (1 << 5)
-    MODE_FLAG_HAS_MODE_SPECIFIC_COLOR = (1 << 6)
-    MODE_FLAG_HAS_RANDOM_COLOR = (1 << 7)
+    HAS_SPEED = (1 << 0)
+    HAS_DIRECTION_LR = (1 << 1)
+    HAS_DIRECTION_UD = (1 << 2)
+    HAS_DIRECTION_HV = (1 << 3)
+    HAS_BRIGHTNESS = (1 << 4)
+    HAS_PER_LED_COLOR = (1 << 5)
+    HAS_MODE_SPECIFIC_COLOR = (1 << 6)
+    HAS_RANDOM_COLOR = (1 << 7)
 
 
 class ModeDirections(IntEnum):
-    MODE_DIRECTION_LEFT = 0
-    MODE_DIRECTION_RIGHT = 1
-    MODE_DIRECTION_UP = 2
-    MODE_DIRECTION_DOWN = 3
-    MODE_DIRECTION_HORIZONTAL = 4
-    MODE_DIRECTION_VERTICAL = 5
+    LEFT = 0
+    RIGHT = 1
+    UP = 2
+    DOWN = 3
+    HORIZONTAL = 4
+    VERTICAL = 5
 
 
 class ModeColors(IntEnum):
-    MODE_COLORS_NONE = 0
-    MODE_COLORS_PER_LED = 1
-    MODE_COLORS_MODE_SPECIFIC = 2
-    MODE_COLORS_RANDOM = 3
+    NONE = 0
+    PER_LED = 1
+    MODE_SPECIFIC = 2
+    RANDOM = 3
 
 
 class DeviceType(IntEnum):
-    DEVICE_TYPE_MOTHERBOARD = 0
-    DEVICE_TYPE_DRAM = 1
-    DEVICE_TYPE_GPU = 2
-    DEVICE_TYPE_COOLER = 3
-    DEVICE_TYPE_LEDSTRIP = 4
-    DEVICE_TYPE_KEYBOARD = 5
-    DEVICE_TYPE_MOUSE = 6
-    DEVICE_TYPE_MOUSEMAT = 7
-    DEVICE_TYPE_HEADSET = 8
-    DEVICE_TYPE_HEADSET_STAND = 9
-    DEVICE_TYPE_UNKNOWN = 10
+    MOTHERBOARD = 0
+    DRAM = 1
+    GPU = 2
+    COOLER = 3
+    LEDSTRIP = 4
+    KEYBOARD = 5
+    MOUSE = 6
+    MOUSEMAT = 7
+    HEADSET = 8
+    HEADSET_STAND = 9
+    UNKNOWN = 10
 
 
 class ZoneType(IntEnum):
-    ZONE_TYPE_SINGLE = 0
-    ZONE_TYPE_LINEAR = 1
-    ZONE_TYPE_MATRIX = 2
+    SINGLE = 0
+    LINEAR = 1
+    MATRIX = 2
 
 
 class PacketType(IntEnum):
-    NET_PACKET_ID_REQUEST_CONTROLLER_COUNT = 0
-    NET_PACKET_ID_REQUEST_CONTROLLER_DATA = 1
-    NET_PACKET_ID_SET_CLIENT_NAME = 50
-    NET_PACKET_ID_RGBCONTROLLER_RESIZEZONE = 1000
-    NET_PACKET_ID_RGBCONTROLLER_UPDATELEDS = 1050
-    NET_PACKET_ID_RGBCONTROLLER_UPDATEZONELEDS = 1051
-    NET_PACKET_ID_RGBCONTROLLER_UPDATESINGLELED = 1052
-    NET_PACKET_ID_RGBCONTROLLER_SETCUSTOMMODE = 1100
-    NET_PACKET_ID_RGBCONTROLLER_UPDATEMODE = 1101
+    REQUEST_CONTROLLER_COUNT = 0
+    REQUEST_CONTROLLER_DATA = 1
+    SET_CLIENT_NAME = 50
+    RGBCONTROLLER_RESIZEZONE = 1000
+    RGBCONTROLLER_UPDATELEDS = 1050
+    RGBCONTROLLER_UPDATEZONELEDS = 1051
+    RGBCONTROLLER_UPDATESINGLELED = 1052
+    RGBCONTROLLER_SETCUSTOMMODE = 1100
+    RGBCONTROLLER_UPDATEMODE = 1101
 
 
 class OpenRGBDisconnected(ConnectionError):
     pass
 
 
-CT = TypeVar("CT", bound="RGBColor") # rgbColor Type
-MDT = TypeVar("MDT", bound="ModeData") # ModeData Type
-ZDT = TypeVar("ZDT", bound="ZoneData") # ZoneData Type
-LDT = TypeVar("LDT", bound="LEDData") # LedData Type
-MEDT = TypeVar("MEDT", bound="MetaData") # MEtaData Type
-CDT = TypeVar("CDT", bound="ControllerData") # ControllerData Type
-PT = TypeVar("PT", bound="Profile") # Profile Type
-
-
 def parse_string(data: bytes, start: int = 0) -> Tuple[int, str]:
     '''
     Parses a string based on a size.
@@ -133,7 +127,7 @@ def pack_list(things: list) -> bytearray:
 
 
 @dataclass
-class RGBColor(object):
+class RGBColor:
     red: int
     green: int
     blue: int
@@ -147,19 +141,15 @@ class RGBColor(object):
         return struct.pack("BBBx", self.red, self.green, self.blue)
 
     @classmethod
-    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, CT]:
+    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, RGBColor]:
         '''
         Unpacks an RGBColor object from bytes
 
         :returns: an RGBColor object
         '''
         size = struct.calcsize("BBBx")
-        if start == 0:
-            r, g, b = struct.unpack("BBBx", data[:size])
-            return size, cls(r, g, b)
-        else:
-            r, g, b = struct.unpack("BBBx", data[start:start + size])
-            return (start + size), cls(r, g, b)
+        r, g, b = struct.unpack("BBBx", data[start:start + size])
+        return (start + size), cls(r, g, b)
 
     @classmethod
     def fromHSV(cls, hue: int, saturation: int, value: int) -> CT:
@@ -170,7 +160,7 @@ class RGBColor(object):
 
 
 @dataclass
-class LEDData(object):
+class LEDData:
     name: str
     value: int
 
@@ -186,7 +176,7 @@ class LEDData(object):
         )
 
     @classmethod
-    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, LDT]:
+    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, LEDData]:
         '''
         Creates a new LEDData object from raw bytes
 
@@ -200,7 +190,7 @@ class LEDData(object):
 
 
 @dataclass
-class ModeData(object):
+class ModeData:
     id: int
     name: str
     value: int
@@ -214,12 +204,34 @@ class ModeData(object):
     color_mode: ModeColors
     colors: List[RGBColor]
 
+    def validate(self):
+        '''
+        Tests the values of the mode data and raises a `ValueError` if the validation fails
+        '''
+        try:
+            if ModeFlags.HAS_SPEED in self.flags:
+                assert self.speed is not None
+                assert self.speed_min <= self.speed <= self.speed_max or self.speed_max <= self.speed <= self.speed_min
+            if ModeFlags.HAS_MODE_SPECIFIC_COLOR in self.flags:
+                assert self.colors_min <= len(self.colors) <= self.colors_max
+        except AssertionError as e:
+            raise ValueError("Mode validation failed.  Required values invalid or not present") from e
+
+        try:
+            if ModeFlags.HAS_SPEED not in self.flags:
+                assert all((i is None for i in (self.speed_max, self.speed_min, self.speed)))
+            if ModeFlags.HAS_MODE_SPECIFIC_COLOR not in self.flags:
+                assert all((i is None for i in (self.colors_max, self.colors_min, self.colors)))
+        except AssertionError as e:
+            raise ValueError("Mode validation failed.  Values are set that are not supported by this mode") from e
+
     def pack(self) -> bytearray:
         '''
         Packs itself into a bytearray ready to be sent to the SDK or saved in a profile
 
         :returns: raw data ready to be sent or saved
         '''
+        self.validate()
         data = (
             struct.pack("i", self.id)
             + pack_string(self.name)
@@ -236,12 +248,12 @@ class ModeData(object):
                 self.color_mode
             )
         )
-        data += pack_list(self.colors)
+        data += pack_list(self.colors if self.colors is not None else [])
         data = struct.pack("I", len(data) + struct.calcsize("I")) + data
         return data
 
     @classmethod
-    def unpack(cls, data: bytearray, start: int = 0, index: int = 0) -> Tuple[int, MDT]:
+    def unpack(cls, data: bytearray, start: int = 0, index: int = 0) -> Tuple[int, ModeData]:
         '''
         Creates a new ModeData object from raw bytes
 
@@ -256,26 +268,26 @@ class ModeData(object):
         buff[1] = ModeFlags(buff[1])
 
         # Garbage data will be sent if these flags aren't set
-        if (ModeFlags.MODE_FLAG_HAS_DIRECTION_HV in buff[1]
-                or ModeFlags.MODE_FLAG_HAS_DIRECTION_UD in buff[1]
-                or ModeFlags.MODE_FLAG_HAS_DIRECTION_LR in buff[1]):
+        if (ModeFlags.HAS_DIRECTION_HV in buff[1]
+                or ModeFlags.HAS_DIRECTION_UD in buff[1]
+                or ModeFlags.HAS_DIRECTION_LR in buff[1]):
             buff[7] = ModeDirections(buff[7])
         else:
             buff[7] = None
-        if ModeFlags.MODE_FLAG_HAS_SPEED not in buff[1]:
+        if ModeFlags.HAS_SPEED not in buff[1]:
             buff[2], buff[3], buff[6] = None, None, None
-        if ModeFlags.MODE_FLAG_HAS_BRIGHTNESS not in buff[1]:
-            buff[4], buff[5] = None, None
-
         buff[8] = ModeColors(buff[8])
-        for i in range(buff[-1]):
-            start, color = RGBColor.unpack(data, start)
-            colors.append(color)
+        if ModeFlags.HAS_MODE_SPECIFIC_COLOR in buff[1]:
+            for i in range(buff[-1]):
+                start, color = RGBColor.unpack(data, start)
+                colors.append(color)
+        else:
+            colors, buff[4], buff[5] = None, None, None
         return start, cls(index, val, *buff[:9], colors)
 
 
 @dataclass
-class ZoneData(object):
+class ZoneData:
     name: str
     zone_type: ZoneType
     leds_min: int
@@ -319,7 +331,7 @@ class ZoneData(object):
         return data
 
     @classmethod
-    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, ZDT]:
+    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, ZoneData]:
         '''
         Unpacks the raw data into a ZoneData object
 
@@ -331,7 +343,7 @@ class ZoneData(object):
         start += struct.calcsize("iIIIH")
         height, width = 0, 0
         matrix = [[]]
-        if buff[0] == ZoneType.ZONE_TYPE_MATRIX:
+        if buff[0] == ZoneType.MATRIX:
             height, width = struct.unpack("II", data[start:start + struct.calcsize("II")])
             start += struct.calcsize("II")
             matrix = [[] for x in range(height)]
@@ -344,7 +356,7 @@ class ZoneData(object):
 
 
 @dataclass
-class MetaData(object):
+class MetaData:
     description: str
     version: str
     serial: str
@@ -364,7 +376,7 @@ class MetaData(object):
         )
 
     @classmethod
-    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, MEDT]:
+    def unpack(cls, data: bytearray, start: int = 0, *args) -> Tuple[int, MetaData]:
         '''
         Unpacks the raw data into a MetaData object
 
@@ -379,7 +391,7 @@ class MetaData(object):
 
 
 @dataclass
-class ControllerData(object):
+class ControllerData:
     name: str
     metadata: MetaData
     device_type: DeviceType
@@ -410,7 +422,7 @@ class ControllerData(object):
         return buff
 
     @classmethod
-    def unpack(cls, data: bytearray, start: int = 0) -> CDT:
+    def unpack(cls, data: bytearray, start: int = 0) -> ControllerData:
         '''
         Unpacks the raw bytes received from the SDK into a ControllerData dataclass
 
@@ -463,7 +475,7 @@ class ControllerData(object):
 
 
 @dataclass
-class Profile(object):
+class Profile:
     '''
     A dataclass to load, store, and pack the data found in an OpenRGB profile file.
     '''
@@ -477,7 +489,7 @@ class Profile(object):
         return data
 
     @classmethod
-    def unpack(cls, profile: BinaryIO) -> PT:
+    def unpack(cls, profile: BinaryIO) -> Profile:
         header = profile.read(16 + struct.calcsize("I"))
         if struct.unpack("16s", header[:16])[0] != b"OPENRGB_PROFILE\x00":
             raise ValueError("The file is not an OpenRGB profile")
@@ -495,19 +507,12 @@ class Profile(object):
             return cls(controllers)
 
 
-class RGBContainer(object):
+class RGBObject:
     '''
     A parent class that includes a few generic functions that use the
     implementation provided by the children.
     '''
 
-    def __init__(self, comms, name: str, device_id: int):
-        self.comms = comms
-        self.name = name
-        self.device_id = device_id
-        self.colors = []
-        self._colors = self.colors[:]
-
     def __repr__(self):
         return f"{type(self).__name__}(name={self.name}, id={self.id})"
 
@@ -531,9 +536,29 @@ class RGBContainer(object):
         '''
         self.clear()
 
+    def update(self):
+        '''
+        Gets the current status from the SDK server, ensuring a correct
+        internal state.
+        '''
+        self.comms.requestDeviceData(self.device_id)
 
-class RGBObject(RGBContainer):
-    def show(self, fast: bool = True, force: bool = False):
+
+class RGBContainer(RGBObject):
+    '''
+    A parent class for RGBObjects that can control more than one LED like the
+    :any:`Device` class or the :any:`Zone` class.
+    '''
+
+    def set_colors(self, colors: List[RGBColor], start: int = 0, end: int = 0, fast: bool = False):
+        '''
+        Sets mutliple colors
+
+        :param colors: A list of colors to set
+        '''
+        pass
+
+    def show(self, fast: bool = False, force: bool = False):
         '''
         Applies changes in the color attribute
         '''
@@ -546,15 +571,12 @@ class RGBObject(RGBContainer):
             self.set_colors(self.colors, fast=True)
         elif len(changed) == 0:
             return
-        elif len(changed) == 1 and not force:
+        elif len(changed) == 1:
             self.leds[changed[0][0]].set_color(changed[0][1], fast=True)
-        elif len(changed) > 1 and not force:
+        elif len(changed) > 1:
             start, end = changed[0][0], changed[-1][0] + 1
-            colors = self.colors[start:end]
+            colors = [color for i, color in changed]
             self.set_colors(colors, start, end, fast=True)
         self._colors = self.colors[:]
         if not fast:
             self.update()
-
-    def update(self):
-        self.comms.requestDeviceData(self.device_id)
